"""
				The softcore.  Persistent words, constants and variables are loaded
				here at startup and remain until the program is terminated.

				Equivalent to a library in other languages.
"""
: constant create , does> @ ;		# Constant creation boilerplate
" [0] " constant fset			# Field set in table used for field count
" [1] " constant dset			# Data set in table used for record count
" [1, " constant lslice
" , " constant comma
" [ " constant lbracket
" ] " constant rbracket
" = " str strip constant equals
fset " ['\n'] " eval slice constant nl	# Python new line character e.g. " This is " nl + " a test. " + .
fset " ['\t'] " eval slice constant tab	# Python tab character
fset " [' '] " eval slice constant space
" None " str strip constant None 
" Junk " str strip constant junk 
" Empty recordset " constant emptyrecordset

: var create 1 allot ;  		# Variable creation boilerplate
var recordset				# working recordset

#	Break out of a loop
""" LoopTotal LoopCount -- LoopCount """ 
: break drop dup 1 - ;

#	Bottom of an ascending loop
""" LoopCount -- True or False """
: next 1 + over over - 0 =  ;

#	Field count and Record count in a recordset
""" recordset -- count """ 
: setFsCount fset recordset @ slice len ;
: setDsCount dset recordset @ slice len ;

#	Formatting which can be placed around cells
: cellBefore space .  ;
# Insert a tab after each field.  cellAfter could be redefined to something else later
""" number -- """
: cellAfter dup 1 + setFsCount - if tab . then ;
#
#  The parent table can have the Join Id field positioned anywhere in the record.
#
# Parent table definition
var TableOuterLoop				# used for parent child tables
var oldTableOuterLoop
var colPos					# colPos is column number in the parent table
var rowPos 					# rowPos is row number in the parent table
var rowPrint					# flag to print joined child row
var skipChildTable				# break out of child table
var skipChildRow				# break out of child row
var landOnRecord 0 landOnRecord !		# next record to begin looking for a join in the child table.
						# Both recordsets MUST be ordered by their Join Id

: parentRow lslice swap dup rowPos ! str + comma + ;
: childRow  lslice swap str + comma + ;
: parentCol swap dup colPos ! str + rbracket + clean recordset @ slice ;
: childCol swap str + rbracket + clean recordset @ slice ;
: nestedTable ;					# used for parent child tables
: parentFilter cellBefore TableOuterLoop @ parentCol . cellAfter ;

: table
setDsCount dup 0 = if drop emptyrecordset . /n else
    0 begin
    	    0 skipChildTable !
	    dup parentRow TableOuterLoop !			# save to a variable to avoid crazy stack manipulations
	    setFsCount 0 begin 
	    	      0 rowPrint !
		      dup parentFilter nestedTable 	 	# nestedTable needs to be removed from
	    next until drop drop /n				# the child table to avoid recursion
    next until drop drop /n 

then
;
#
# The child table must ALWAYS have the Join Id as the FIRST field!
#
# Child table definition
: cellFormatted cellBefore TableOuterLoop @ childCol . cellAfter ;
: childFilter
   cellFormatted # The default filter is a Cartesian join.
;

: childTable
    TableOuterLoop @ oldTableOuterLoop !
    setDsCount dup 0 = if drop emptyrecordset . /n else
      landOnRecord @ 0 = if 0 else landOnRecord @ then
      begin
	      dup childRow TableOuterLoop !		# save to a variable to avoid crazy stack manipulations
	      rowPrint @ 0 = if else /n then		# newLine at beginning of child record to be printed
    	      0 skipChildRow !
	      setFsCount 0 begin 
		      dup childFilter # Whole field placed in a single word so that a filter can be applied to all of it		    
		      skipChildRow @ 0 = if skipChildTable @ 0 = if else break then else break then
	      next until drop drop 
	      skipChildTable @ 0 = if else dup landOnRecord ! break then # remember row to start on next time 
      next until drop drop /n 
    then
    oldTableOuterLoop @ TableOuterLoop !
;
